import sys
import click
import logging
from neoload.neoload_cli_lib import cli_exception, tools
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory
import json
import re
import urllib.parse
import requests
import io

import os
import os.path
from base64 import b64encode

@click.group(chain=True)
@CommandCategory("Converting")
def cli():
    """Configure this utility to simplify some commands
    """
    pass

wrap_requests_in_transactions="wrap_requests_in_transactions"
delay_in_seconds="delay_in_seconds"
append_headers="append_headers"
add_javascript_lib_postman="add_javascript_lib_postman"
javascript_lib_postman_var_name='postman'

@cli.command('postman')
@click.option("--file", "-f", help="The Postman file to import")
@click.option("--filter", required=True, help="A filter statement to match key/value pairs")
@click.option("--json", 'output_to_json', is_flag=True, default=False, help="Output internal JSON instead of YAML")
@click.option("--grouping", default='NONE', type=click.Choice(['NONE', 'TRANSACTION']), help="How requests should be grouped; Default=NONE means each request is wrapped in a separate transaction under it's appropriate parent transaction")
@click.option("--delay", default="1s", help="Delay between HTTP requests")
def postman(file, filter, output_to_json, grouping, delay):
    """EXAMPLE: nlc import postman --filter='name=GET Request' -f exported_collection.json
    """
    if not (file and len(file) > 0):
        raise cli_exception.CliException("You must use the --file option")

    data = None

    if is_url(file):
        logging.debug("Import file is a URL")
        resp = requests.get(file, stream=True)
        data = json.load(io.BytesIO(resp.content))
    else:
        file = os.path.expanduser(file)
        if os.path.exists(file):
            with open(file) as stm:
              data = json.load(stm)
        else:
            logging.debug("Import file does not exist")

    if data is None:
        tools.system_exit({'code':1,'message':'Could not determine data to import'})
        return

    filter_parsed = parse_filter_spec(filter)
    build = builder_data.BuilderData()
    options = {}
    options[wrap_requests_in_transactions] = (grouping == "NONE")
    options[delay_in_seconds] = delay
    options[add_javascript_lib_postman] = False

    version = None

    if 'info' in data and 'schema' in data['info']:
        schema = data['info']['schema']
        version = re.search(r"/v(.*?)/", schema).group(1)

    if version and version.startswith("2."):
        #add_headers(build, get_auth_headers(data), all=True)
        process_postman_v2(build, data, filter_parsed, options)

    if options[add_javascript_lib_postman] == True:
        build.add(builder_data.VariableTypeJavascript(
            name=javascript_lib_postman_var_name,
            description='A javascript library for running postman tests using the Node sidecar',
            script=get_postman_var_script_source(),
            change_policy='each_use'
        ))


    builder_data.save(build)
    output = build if output_to_json else builder_data.convert_builder_to_yaml(build)
    #output = builder_data.get() if output_to_json else builder_data.convert_builder_to_yaml(builder_data.get())
    print(output)

def add_headers(build, headers, all=False):
    for key in headers:
        build.add(builder_data.Header(name=key, value=",".join(headers[key]), all=all))


def is_url(url):
    try:
    # python 3
        from urllib.parse import urlparse
    except ImportError:
        from urlparse import urlparse
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except:
        return False

def process_postman_v2(build, data, filter, options):
    process_postman_items(build, data, filter, options)

def process_postman_items(build, parent, filter, options):
    if 'item' in parent:
        for item in parent['item']:
            item['__parent'] = parent

            matches = item_matches_filter(item, filter) or item_matches_filter(parent, filter)

            if matches:
                if not ('__added' in parent and parent['__added'] == True):
                    process_postman_item(build, parent['__parent'] if '__parent' in parent else None, parent, options)

                process_postman_item(build, parent, item, options)
            # else:
            #     print('Does not match: {}'.format(item['name']))

            process_postman_items(build, item, filter, options)


def item_matches_filter(item, filter):
    if len(filter) > 0:
        for key in filter: # filters may be more than one value
            matches = filter[key]
            ret = False # assume not a match unless at least one value matches
            for match in matches:
                if key in item: # some elements don't have the filter key
                    val = item[key]
                    if val == match:
                        ret = True
        return ret
    else:
        return True # no filter means everything

def is_postman_item_parent(item):
    if 'name' in item and not ('request' in item):
        return True
    else:
        return False

def process_postman_item(build, parent, item, options):
    item['__added'] = True
    if 'name' in item:
        name = item['name']
        description = item['description'] if 'description' in item else None
        #print("Found item: {}".format(name))
        parent_name = parent['name'] if 'name' in parent else 'parent'
        if 'request' in item:
            request = item['request']
            details = {
                'method': "{}".format(request['method']),
                'url': "{}".format(request['url']['raw'] if type(request) is dict and 'raw' in request['url'] else request['url']),
                'body': "{}".format(parse_postman_body(request)),
                'description': "{}".format(description)
            }

            if options[wrap_requests_in_transactions]:
                build.add(builder_data.Transaction(name=name, description=description, inside=parent_name))

            build.add(builder_data.HttpRequest(details))

            add_headers(build, get_auth_headers(item))

            if 'event' in item:
                logging.debug("events")
                for event in item['event']:
                    logging.debug("event: {}".format(event))
                    if event['listen'] == "test" and 'script' in event:
                        script = event['script']
                        if 'type' in script and script['type'] == "text/javascript" and 'exec' in script:
                            options[add_javascript_lib_postman] = True
                            build.add(builder_data.JavascriptAction(
                                name="Postman Tests",description="",
                                script=get_postman_action_script_source(script['exec'])
                            ))

            build.add(builder_data.Delay(time=options[delay_in_seconds]))

            if 'header' in request:
                for header in request['header']:
                    build.add(builder_data.Header(name=header['key'], value=header['value'], all=False))

        elif is_postman_item_parent(item):
            build.add(builder_data.Transaction(name=name, description=description, inside=parent_name))

        else:
            print("Unparseable item")


def parse_postman_body(item):
    if 'body' in item:
        body = item['body']
        mode = body['mode']
        if mode == "raw":
            return body[mode]
        elif mode == "urlencoded":
            return parse_urlencoded_body(body[mode])
        else:
            print("Could not parse body type of '{}'".format(mode))

    return None

def parse_urlencoded_body(pairs):
    return "&".join(list(map(lambda x: "{}={}".format(urllib.parse.quote(x['key']),urllib.parse.quote(x['value'])),pairs)))

def parse_filter_spec(filter_spec):
    ret = {}

    if filter_spec is not None and len(filter_spec)>0:
        filter_parts = filter_spec.split("|" if "|" in filter_spec else ";")
        for part in filter_parts:
            subparts = part.split("=")
            key = subparts[0]
            value = "=".join(subparts[1:])
            if key not in ret:
                ret[key] = []

            ret[key].append(value)

    return ret

def get_auth_headers(item, recursion_level=0):
    ret = {}
    #logging.debug("[{}]get_auth_headers::recursion_level = {}".format(item['name'] if 'name' in item else 'Anonymous Object',recursion_level))

    inherit_auth = 'auth' not in item
    if inherit_auth:
        ret.update(get_auth_headers(item['__parent'], recursion_level=(recursion_level+1)))
    else:
        auth = item['auth']
        auth_type = auth['type']
        if auth_type == "noauth":
            return ret
        elif auth_type == "apikey":
            apikey = auth[auth_type]
            values = single_or_list(lambda x: (x['key'],x['value']),apikey)
            for (key,value) in values:
                if key not in ret:
                    ret[key] = []
                ret[key].append(value)
        elif auth_type == "bearer":
            bearer = auth[auth_type]
            ret["Authorization"] = ["Bearer {}".format(bearer['token'])]
        elif auth_type == "basic":
            username = auth[auth_type]['username']
            password = auth[auth_type]['password']
            encoded_credentials = b64encode(bytes(f'{username}:{password}',
                                encoding='ascii')).decode('ascii')
            auth_header = f'Basic {encoded_credentials}'
            ret["Authorization"] = [auth_header]
        else:
            logging.warn("Unsupported auth type '{}'".format(auth_type))

    return ret

def single_or_list(fun,node):
    if type(node) is list:
        return list(map(fun,node))
    else:
        return [fun(node)]

def get_postman_var_script_source():
    return """
function evaluate(){var t={};return new function(){this.static=get_postman_results_(t)}}function get_postman_results_(t){var e=context,s=e.variableManager,r=function(t){logger.debug(t)},a=java.nio.charset.Charset.defaultCharset(),o=function(t){e.fail(t)},n="get_postman_result_",u=s.getValue("__response___json")+"",i=s.getValue("__postman___tests")+"";u||o("no __response__ data"),i||o("no __postman___tests data");var c=u+"---CONTEXT---"+JSON.stringify({tests:i,pm:t});r(n+"[stdin]: "+c);var l=new java.lang.ProcessBuilder("/usr/local/bin/node","/nlpm/nlpm-exec.js").start(),_=new java.io.BufferedWriter(new java.io.OutputStreamWriter(l.getOutputStream(),"UTF-8"));_.write(c),_.flush(),_.close();var g=org.apache.commons.io.IOUtils.toString(l.getErrorStream(),a),p=org.apache.commons.io.IOUtils.toString(l.getInputStream(),a);r(n+"[stdout]: "+p),r(n+"[stderr]: "+g),g&&null!=g&&g.length>0&&o(n+"[stderr]: "+g);var m={success:!1};try{var d=JSON.parse(p);if(m.context=d,m.success=1==d.summary.all,!m.success)throw"Not all tests passed."}catch(t){m.error=t}return JSON.stringify(m)}
    """

def get_postman_action_script_source(tests_lines):
    pmscript_encoded = urllib.parse.quote("\n".join(tests_lines))
    action_script = """
var e = '{urlencoded}'
var c = context; var vm = c.variableManager; var j = JSON; vm.setValue("__postman___tests",e); var r = j.parse(vm.getValue("{javascript_lib_postman_var_name}.static"))
if(!(r && r.success && r.success == true)) c.fail(r.error +': ' + j.stringify(r, null, '\t'))
    """
    return action_script.format(urlencoded=pmscript_encoded,javascript_lib_postman_var_name=javascript_lib_postman_var_name)



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.argument("duration", type=str)
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, duration):
    """Adds a delay (think time) to the builder queue
    """
    builder_data.register_context(ctx)

    if duration is None:
        raise cli_exception.CliException("You must provide a time for this delay")

    builder = builder_data.get() \
        .add(builder_data.Delay(time=duration)) \
        .save()



import sys
import click

from compose_lib import builder_data, profile, common
from commands import config
import tempfile
import yaml
import json
from compose_lib.common import os_run, os_return
from compose_lib.command_category import CommandCategory

from neoload.neoload_cli_lib import tools

@click.command()
@click.argument("name_or_id", required=False)
@click.option('--zone', required=False, help="The zone to run this test in.")
@click.option('--scenario', required=False, help="The scenario to run.")
@click.option('--save', is_flag=True, help="Save this as the default zone and test-setting")
@click.option('--just-report-last', is_flag=True, default=False, help="Save this as the default zone and test-setting")
@click.pass_context
@CommandCategory("Validating")
def cli(ctx, name_or_id, zone, scenario, save, just_report_last):
    """Runs whatever is in the current buffer
    """
    builder_data.register_context(ctx, auto_reset=False)

    neoload_base_cmd = "neoload " + ("--debug " if common.get_debug() else "")

    if not just_report_last:
        if name_or_id and save:
            config.test_setting(name_or_id)
        if zone and save:
            config.zone(zone)

        if not name_or_id:
            name_or_id = profile.get().default_test_setting
            if not name_or_id:
                tools.system_exit({'code':3,'message':"No test settings [name_or_id] provided and no default test-setting configured!"})
                return

        if not zone:
            zone = profile.get().default_zone

        if zone == "any":
            proc = os_return(neoload_base_cmd + " zones", status=False)
            (stdout,strerr) = proc.communicate()
            result = json.loads(stdout)
            availables = list(filter(lambda z: any(filter(lambda c: c['status'] == "AVAILABLE",z['controllers'])) and any(filter(lambda lg: lg['status'] == "AVAILABLE",z['loadgenerators'])),result))
            if len(availables) > 0:
                availables = sorted(availables, key=lambda z: 0 if z['type'] == "STATIC" else 1)
                zone = availables[0]['id']
                print("Because zone is 'any', the zone '{}' ({}) has been automatically selected.".format(zone, availables[0]['name']))
            else:
                tools.system_exit({'code':3,'message':"There are no zones with available load generators and controllers!!!"})
                return

        if not zone:
            tools.system_exit({'code':3,'message':"No --zone provided and no default zone configured!"})
            return

        if not os_run(neoload_base_cmd + " status", status=False):
            return

        yaml_str = builder_data.convert_builder_to_yaml(builder_data.get())
        data = yaml.safe_load(yaml_str)

        if not scenario:
            scenario = data['scenarios'][0]['name']

        dir = None
        with tempfile.TemporaryDirectory() as tmp:
            dir = tmp
            builder_data.write_to_path(dir, yaml_str)

            if not os_run(neoload_base_cmd + " test-settings --zone {} --lgs 1 --scenario {} createorpatch {}".format(zone, scenario, name_or_id), status=True):
                return

            if not os_run(neoload_base_cmd + " project --path {} up".format(dir), status=True):
                return


        if not os_run(neoload_base_cmd + " run".format(scenario), print_stdout=True, print_line_check=check_run_line):
            print("Test failed.")


    proc = os_return(neoload_base_cmd + " report --help", status=False)
    (stdout,strerr) = proc.communicate()
    if proc.returncode != 0 or 'Error:' in stdout.decode("UTF-8"):
        print("Test ran, but could not produce final (pretty) report. {}".format("" if strerr is None else strerr))
    else:
        import pkg_resources

        template = pkg_resources.resource_filename(__name__, 'resources/dist/jinja/builtin-console-summary.j2')

        if not os_run(neoload_base_cmd + " report --template {} --filter '{}' cur".format(
                    template,
                    'exclude=events,slas,all_requests,ext_data,controller_points'
                ),
                status=True,
                print_stdout=True):
            return

__pause_output = False
def check_run_line(line_text):
    global __pause_output
    if 'SLA summary:' in line_text:
        __pause_output = True
    if __pause_output and line_text.startswith('Test '):
        __pause_output = False
    return not __pause_output



import sys
import click

from compose_lib.command_category import CommandCategory

@click.command()
def cli():
    """Provides simple examples of usage
    """
    print("""


nlc transaction GetAndPost \\
    http --get http://httpbin.org/get?test=123 delay 1s \\
    extract --name traceId \\
            --jsonpath ".headers['X-Amzn-Trace-Id']" \\
            --regexp "=(.*)" \\
    http --post http://httpbin.org/post --body "{'trace_id':'${traceId}'}" delay 1s \\
    sla --name PostSLA per-interval --error-rate --warn ">= 10%" --fail ">= 20%"


nlc import postman --filter="name=Request Methods" \\
 -f "{{repo_content_base}}/doc/examples/example_postman_collection_export.json"



    """.replace("{{repo_content_base}}","https://raw.githubusercontent.com/paulsbruce/neoload-compose/master"))



import sys
import click
import ruamel.yaml
import json

from compose_lib import profile, builder_data
from compose_lib.command_category import CommandCategory
import tempfile
import subprocess

yaml = ruamel.yaml.YAML(typ='unsafe')

@click.group(chain=True)
@CommandCategory("Configuration")
def cli():
    """Configure this utility to simplify some commands
    """
    pass

@cli.command('reset')
@click.option("--confirm", is_flag=True, default=False, help="Must confirm that you want to reset the profile data")
def reset(confirm):
    """Resets the current configuration
    """
    if confirm != True:
        raise cli_exception.CliException("You must use the --confirm option")

    profile.reset()

@cli.command('current')
def current():
    """Prints the current configuration
    """
    yaml.register_class(profile.ProfileData)
    #yaml.register_class(dict)
    builder_data.register_classes(yaml)
    data = {
        'profile': profile.get(),
        'builder': {
            'filepath': builder_data.get_storage_filepath(),
            #'data': builder_data.get()
        }
    }
    yaml.dump(data, sys.stdout)


@cli.command('zone')
@click.argument("name_or_id")
def zone(name_or_id):
    """Sets the default zone to use in zone-related subcommands; can also be 'any'
    """
    profile.get() \
        .set_default_zone(name_or_id) \
        .save()


@cli.command('test-setting')
@click.argument("name_or_id")
def test_setting(name_or_id):
    """Set the default test-setting to use for instant run and other subcommands
    """
    profile.get() \
        .set_default_test_setting(name_or_id) \
        .save()



import sys
import click

from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

import os

@click.command()
@click.option("--json", 'output_to_json', is_flag=True, default=False, help="Output internal JSON instead of YAML")
@click.option("--path", help="Write current test to a file or folder")
@click.pass_context
@CommandCategory("Validating")
def cli(ctx, output_to_json, path):
    """Prints out the current builder state
    """
    builder_data.register_context(ctx, auto_reset=False)

    output = builder_data.get() if output_to_json else builder_data.convert_builder_to_yaml(builder_data.get())

    if path is not None:
        builder_data.write_to_path(path, output)

    else:
        print(output)



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.argument("name", type=str)
@click.option('--description','-d', help="The description of this transaction")
@click.option("--inside", default='last', help="Append this transaction inside a prior transaction")
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, name, description, inside):
    """Adds a transaction to the builder queue; do before requests
    """
    builder_data.register_context(ctx)

    if name is None:
        raise cli_exception.CliException("You must provide a name for this transaction")

    builder = builder_data.get() \
        .add(builder_data.Transaction(name=name, description=description, inside=inside)) \
        .save()



import sys
import click

from neoload.neoload_cli_lib import cli_exception, tools
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

current_name = None

# nlc -c sla --name geo_3rd_party per-interval --error-rate --warn '>= 5s' --fail '>= 10s'
#
# avg-request-resp-time   	Average request response time	s or ms
# avg-page-resp-time  	Average page response time	s or ms
# avg-transaction-resp-time   	Average Transaction response time	s or ms
# perc-transaction-resp-time  	Percentile Transaction response time
# avg-request-per-sec 	Average requests per second	/s
# avg-throughput-per-sec  	Average throughput per second	Mbps
# errors-count	   Total errors	-
# count   	Total count	-
# throughput	 Total throughput	MB
# error-rate  	Error rate	%
#
# avg-resp-time   	Average response time	s or ms
# avg-elt-per-sec	    Average elements per second	/s
# avg-throughput-per-sec	 Average throughput per second	Mbps
# errors-per-sec	 Errors per second	/s
# error-rate	 Error rate	%

@click.group(chain=True)
@click.option("--name", required=True, help="Name of the SLA profile")
@CommandCategory("Composing")
def cli(name):
    """Create and apply an SLA of type per-interval or per-test
    """
    global current_name
    current_name = name

@cli.command('per-interval')
@click.option("--avg-resp-time", is_flag=True, default=False, help="Average response time	s or ms")
@click.option("--avg-elt-per-sec", is_flag=True, default=False, help="Average elements per second	/s")
@click.option("--avg-throughput-per-sec", is_flag=True, default=False, help="Average throughput per second	Mbps")
@click.option("--errors-per-sec", is_flag=True, default=False, help="Errors per second	/s")
@click.option("--error-rate", is_flag=True, default=False, help="Error rate	%")
@click.option("--warn", help="An operator, value, and unit spec to raise an SLA warning")
@click.option("--fail", help="An operator, value, and unit spec to raise an SLA failure")
def per_interval(avg_resp_time, avg_elt_per_sec, avg_throughput_per_sec, errors_per_sec, error_rate, warn, fail):
    """Create a per-interval threshold
    """

    if not (avg_resp_time or avg_elt_per_sec or avg_throughput_per_sec or errors_per_sec or error_rate):
        tools.system_exit({'code':1,'message':"At least one KPY flag must be provided"})
        return

    if warn is None and fail is None:
        tools.system_exit({'code':1,'message':"Either one or both 'warn' and 'fail' flags must be provided"})
        return

    global current_name

    builder = builder_data.get() \
        .add(builder_data.SLAThreshold(
            name=current_name,
            scope="per interval",
            warn=warn,
            fail=fail,
            kpis={
                'avg-resp-time': avg_resp_time,
                'avg-elt-per-sec': avg_elt_per_sec,
                'avg-throughput-per-sec': avg_throughput_per_sec,
                'errors-per-sec': errors_per_sec,
                'error-rate': error_rate
            }
        )
        ) \
        .save()



import logging
import os
import re
import sys

import json
import ruamel.yaml
import re
import os
import click

from collections import OrderedDict
from ruamel.yaml.comments import CommentedMap as ordereddict
from neoload_cli_lib import cli_exception
from compose_lib import common

from neoload_compose import get_global_continue
from compose_lib import profile

__builder_config_file = os.path.join(common.__config_dir, "builder.yaml")

def reset():
    global __builder_data_singleton
    __builder_data_singleton = None
    if os.path.exists(__builder_config_file):
        os.remove(__builder_config_file)


def get(throw=True):
    global __builder_data_singleton
    if __builder_data_singleton is None:
        init()
    return __builder_data_singleton

def get_storage_filepath():
    return __builder_config_file

def save(build):
    global __builder_data_singleton
    __builder_data_singleton = build
    BuilderData.config_file = __builder_config_file
    build.config_file = __builder_config_file
    build.save()
    return __builder_data_singleton

def init():
    return save(BuilderData())

def d_print(o):
    pass#print(o)

def register_context(ctx, auto_reset=True):
    d_print("dump_context[{}]".format(ctx.info_name))
    if ctx.parent is not None and "NeoLoadCompose" in "{}".format(ctx.parent.command):
        if not 'heirarchy' in ctx.parent.__dict__:
            ctx.parent.__dict__['heirarchy'] = []
        is_first = (len(ctx.parent.__dict__['heirarchy']) == 0)
        ctx.parent.__dict__['heirarchy'].append({
            'position': len(ctx.parent.__dict__['heirarchy']),
            'ctx': ctx
        })
        if is_first and auto_reset and not get_global_continue():
            logging.debug("AUTO-RESET on first command!")
            reset()

    recurse_obj(ctx,0)

def recurse_obj(ctx,iteration):
    inset = "\t" * iteration
    d_print("{}{}".format(inset,ctx))
    for key in ctx.__dict__:
        val = ctx.__dict__[key]
        if iteration < 2 and (
            type(val) is click.core.Context
            or
            type(val) is click.core.Command
        ):
            recurse_obj(val, iteration+1)
        else:
            d_print("{}{}: {}".format(inset,key,val))


class BuilderData(common.StorableConfig):
    def __init__(self):
        super(BuilderData, self).__init__()
        self.stack = []

    def add(self, thing_to_add):
        self.stack.append(thing_to_add)
        return self


class ContainableItem:
    pass

class Action(ContainableItem):
    def __init__(self):
        pass

class HttpRequest(Action):
    def __init__(self, details):
        super(HttpRequest, self).__init__()
        self.details = details
        self.headers = []
        self.extractors = []
        self.sla_profile = None

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'request': common.remove_empty({
                'url': self.details['url'],
                'method': self.details['method'] if self.details['method'] != "GET" else None,
                'body': self.details['body'],
                'description': self.details['description'] if 'description' in self.details else None,
                'headers': self.headers if len(self.headers) > 0 else None,
                'extractors': self.extractors if len(self.extractors) > 0 else None,
                'sla_profile': self.sla_profile if 'sla_profile' in self.__dict__ else None
            })
        }

        return dumper.represent_data(new_data)

class Extractor():
    def __init__(self, name, jsonpath, xpath, regexp, from_, match_number, template, decode, extract_once, default, throw_error):
        self.details = {
            'name': name,
            'jsonpath': jsonpath,
            'xpath': xpath,
            'regexp': regexp,
            'from': from_,
            'match_number': match_number,
            'template': template,
            'decode': decode,
            'extract_once': extract_once,
            'default': default,
            'throw_assertion_error': throw_error
        }

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'name': self.details['name']
        }
        self.add_default(new_data, 'jsonpath', None)
        self.add_default(new_data, 'xpath', None)
        self.add_default(new_data, 'regexp', None)
        self.add_default(new_data, 'from', 'body')
        self.add_default(new_data, 'match_number', 1)
        self.add_default(new_data, 'template', '$1$')
        self.add_default(new_data, 'decode', None)
        self.add_default(new_data, 'extract_once', False)
        self.add_default(new_data, 'default', None)
        self.add_default(new_data, 'throw_assertion_error', True)

        new_data = common.remove_empty(new_data)

        return dumper.represent_data(new_data)

    def add_default(self, to_col, key, defaultValue):
        if key in self.details:
            #to_col[key] = self.details[key] if self.details[key] is not None else defaultValue
            to_col[key] = self.details[key]

class SLAThreshold():
    def __init__(self, name, scope, warn, fail, kpis):
        if not (kpis is not None and len(kpis) > 0):
            raise ValueError("Error in SLAThreshold: no kpi flags to be recorded!")

        self.details = {
            'name': name,
            'scope': scope,
            'warn': warn,
            'fail': fail,
            'kpis': kpis
        }

class SLA():
    def __init__(self, name):
        self.name = name
        self.thresholds = []

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'name': self.name,
            'thresholds': []
        }
        for threshold in self.thresholds:
            str_value = ""
            kpis = threshold.details['kpis']
            for key in kpis.keys():
                if kpis[key] == True:
                    str_value = key
                    if threshold.details['warn'] is not None:
                        str_value += " warn {}".format(threshold.details['warn'])
                    if threshold.details['fail'] is not None:
                        str_value += " fail {}".format(threshold.details['fail'])

                    str_value += " {}".format(threshold.details['scope'])
                    break

            new_data['thresholds'].append(str_value)

        logging.debug(new_data)

        return dumper.represent_data(new_data)

class Header:
    def __init__(self, name, value, all):
        self.name = name
        self.value = value
        self.all = all

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {}
        new_data[self.name] = self.value

        return dumper.represent_data(new_data)


class Delay(Action):
    def __init__(self, time):
        super(Delay, self).__init__()
        self.time = convert_time_to_duration_format(time)

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'delay': self.time
        }

        return dumper.represent_data(new_data)

class JavascriptAction(Action):
    def __init__(self, script, name=None, description=None):
        super(JavascriptAction, self).__init__()
        self.name = name
        self.description = description
        self.script = script

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = common.remove_empty({
            'name': self.name,
            'description': self.description,
            'script': self.script
        })

        return dumper.represent_data(new_data)

class Variable():
    def __init__(self, name, description=None):
        self.name = name
        self.description = description
        pass

class VariableWithChangePolicy(Variable):
    def __init__(self, name, change_policy, description=None):
        super(VariableWithChangePolicy, self).__init__(name=name,description=description)
        self.change_policy = change_policy


class VariableTypeJavascript(VariableWithChangePolicy):
    def __init__(self, name, script, change_policy, description=None):
        super(VariableTypeJavascript, self).__init__(name=name, change_policy=change_policy, description=description)
        self.name = name
        self.description = description
        self.script = script

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'javascript': common.remove_empty({
                'name': self.name,
                'description': self.description,
                'script': self.script,
                'change_policy': self.change_policy
            })
        }

        return dumper.represent_data(new_data)

def convert_time_to_duration_format(in_spec):
    matches = re.finditer("ms|d|h|m|s", in_spec)
    ret = in_spec
    offset = 0
    for m in matches:
        (start,end) = m.span()
        ret = ret[:(offset+end)] + " " + ret[(offset+end):]
        offset += 1

    while("  " in ret):
        ret = ret.strip().replace("  ", " ")

    ret = ret.strip()

    return ret


class Container(ContainableItem):
    def __init__(self, parent=None, name=None, description=None):
        self.parent = parent
        self.name = name
        self.description = description
        self.steps = []
        self.sla_profile = None

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = common.remove_empty({
            'name': self.name,
            'description': self.description,
            'steps': self.steps,
            'sla_profile': self.sla_profile
        })

        return dumper.represent_data(new_data)

class Transaction(Container):
    def __init__(self, parent=None, name=None, description=None, inside='last'):
        super(Transaction, self).__init__(parent=parent, name=name, description=description)
        self.inside = inside

    @classmethod
    def to_yaml(cls,dumper,self):
        new_data = {
            'transaction': common.remove_empty({
                'name': self.name,
                'description': self.description,
                'steps': self.steps,
                'sla_profile': self.sla_profile if 'sla_profile' in self.__dict__ else None
            })
        }

        return dumper.represent_data(new_data)

class UserPath():
    def __init__(self, name=None, description=None):
        self.name = name
        self.description = description
        self.init = Container(self)
        self.actions = Container(self)
        self.end = Container(self)

    @classmethod
    def to_yaml(cls,dumper,self):
        #logging.debug(cls.yaml_flow_style)
        new_data = common.remove_empty(ordereddict({
            'name': self.name,
            'description': self.description,
        }))
        for key in ['init','actions','end']:
            if len(self.__dict__[key].steps) > 0:
                new_data[key] = self.__dict__[key]

        return dumper.represent_data(new_data)

class Population:
    def __init__(self, name=None, description=None):
        self.name = name
        self.description = description
        self.paths = []

class Scenario:
    def __init__(self, name=None, description=None):
        self.name = name
        self.description = description
        self.populations = []

    @classmethod
    def to_yaml(cls,dumper,self):
        #logging.debug(cls.yaml_flow_style)
        scn_data = common.remove_empty(ordereddict({
            'name': self.name,
            'description': self.description,
            'populations': []
        }))
        for pop in self.populations:
            var_pol = pop['variation_policy']
            pop_data = {
                'name': pop['population'].name,
            }
            pop_data.update(var_pol.provide_inner_data())
            scn_data['populations'].append(pop_data)

        return dumper.represent_data(scn_data)

class DurationPolicy():
    def __init__(self, duration):
        self.duration = duration

class VariationPolicy(DurationPolicy):
    def __init__(self, duration=None):
        super(VariationPolicy, self).__init__(duration=duration)

    def provide_inner_data(self): return {}

    @classmethod
    def to_yaml(cls,dumper,self):
        #logging.debug(cls.yaml_flow_style)
        new_data = self.provide_inner_data()

        return dumper.represent_data(new_data)

class ConstantPolicy(VariationPolicy):
    def __init__(self, vus=5, duration="1m"):
        super(ConstantPolicy, self).__init__(duration=duration)
        self.vus = vus

    def provide_inner_data(self):
        ret = {
            'constant_load': {
                'users': self.vus,
                'duration': self.duration
            }
        }
        return ret

class RampPolicy(VariationPolicy):
    def __init__(self, to=5, by=1, per="5s", duration=None):
        super(RampPolicy, self).__init__(duration=duration)
        self.to = to
        self.by = by
        self.per = convert_time_to_duration_format(per)

    def provide_inner_data(self):
        ret = {
            'rampup_load': common.remove_empty(ordereddict({
                'min_users': 1,
                'max_users': self.to,
                'increment_users': self.by,
                'increment_every': self.per,
                'duration': self.duration
            }))
        }
        return ret

def convert_builder_to_yaml(builder):
    user_paths = []
    slas = []

    default_population = Population(name='DefaultPopulation')
    default_scenario = Scenario(name='DefaultScenario')

    current_path = UserPath('UserPath1')
    container_heirarchy = [current_path.actions]
    last_item = container_heirarchy[-1]

    current_population = default_population
    current_scenario = default_scenario
    global_duration = DurationPolicy("1m")
    current_duration = global_duration
    global_variables = []

    all_requests_no_transactions = (
        len(list(filter(lambda item: type(item) is HttpRequest, builder.stack)))>0
        and
        len(list(filter(lambda item: type(item) is Transaction, builder.stack)))<1
    )

    current_request = None
    global_headers = []
    all_slas = []

    temp_transaction_counter = 0

    # recurse for appliables
    for item in builder.stack:

        parent = get_parent_for_item(item, container_heirarchy, current_path)

        if all_requests_no_transactions and type(item) is HttpRequest:
            temp_transaction_counter += 1
            temp_transaction = Transaction(name="Transaction {}".format(temp_transaction_counter), inside=parent.name)
            parent.steps.append(temp_transaction)
            parent = temp_transaction

        if issubclass(type(item), ContainableItem):
            parent.steps.append(item)

        if type(item) is Container or issubclass(type(item), Container):
            container_heirarchy.append(item)

        if issubclass(type(item), VariationPolicy):
            current_scenario.populations.append({'population': current_population,'variation_policy': item})

        if type(item) is Population:
            current_population = item

        if type(item) is Scenario:
            current_scenario = item

        if type(item) is DurationPolicy:
            current_duration = item

        if type(item) is HttpRequest:
            current_request = item

        if type(item) is Header:
            if item.all:
                global_headers.append(item)
            else:
                if current_request is not None: current_request.headers.append(item)

        if type(item) is Extractor:
            if current_request is not None: current_request.extractors.append(item)

        if type(item) is SLAThreshold:
            # add an SLA object to all_slas
            threshold = item
            sla_name = threshold.details['name']
            found_sla = list(filter(lambda sla: sla.name == sla_name, all_slas))
            if len(found_sla) > 0:
                found_sla = found_sla[0]
            else:
                found_sla = SLA(sla_name)
                all_slas.append(found_sla)

            # add this threshold to that object's thresholds
            found_sla.thresholds.append(threshold)

            # apply this named sla to the last_item if request or transaction
            prior_sla_item = current_request
            if prior_sla_item is None:
                prior_sla_item = container_heirarchy[-1]
            if type(prior_sla_item) is Transaction or type(prior_sla_item) is HttpRequest:
                prior_sla_item.sla_profile = found_sla.name

        if type(item) is Variable or issubclass(type(item), Variable):
            global_variables.append(item)

        last_item = item

    for header in global_headers:
        apply_header_to_user_path(header, current_path)

    current_population.paths.append(current_path)
    user_paths.append(current_path)

    # if no explicit policy defined
    if len(current_scenario.populations) < 1:
        current_scenario.populations.append({
            'population': current_population,
            'variation_policy': ConstantPolicy()
        })

    scenarios = [current_scenario]

    for scn in scenarios:
        for holder in scn.populations:
            if holder['variation_policy'].duration is None:
                holder['variation_policy'].duration = current_duration.duration

    project_name = profile.get().default_test_setting
    if not (project_name is not None and len(project_name.strip())):
        project_name = "Project1"

    project = {
        'name': project_name,
        'user_paths': user_paths,
        'populations': [],
        'scenarios': scenarios,
        'sla_profiles': [],
        'variables': global_variables
    }

    if len(project['variables']) < 1:
        del project['variables']

    for scn in scenarios:
        for holder in scn.populations:
            project['populations'].append({
                'name': holder['population'].name,
                'user_paths': list(map(lambda p: { 'name': p.name, 'distribution': '100%' }, holder['population'].paths))
            })

    for sla in all_slas:
        project['sla_profiles'].append(sla)

    if len(project['sla_profiles']) < 1:
        del project['sla_profiles']

    yaml = ruamel.yaml.YAML()

    yaml.encoding = None
    register_classes(yaml)

    fun = MyLogger()
    yaml.encoding = None
    yaml.dump(project, fun, transform=strip_python_tags)
    return fun.readAll()

def get_parent_for_item(item, container_heirarchy, current_path):
    parent = container_heirarchy[-1]

    # solves for where to stick this item, based on 'inside' spec
    if type(item) is Container or issubclass(type(item), Container):
        #print('item.name={}'.format(item.name))
        #print(container_heirarchy)
        if item.inside == 'last' and container_heirarchy[-1] != current_path.actions:
            #container_heirarchy.pop()
            parent = container_heirarchy[-1]
        elif type(item) is Transaction:
            if item.inside == 'parent':
                for i in range(1,len(container_heirarchy)+1):
                    #print('ha: '.format(container_heirarchy[-i].name))
                    if not ('inside' in container_heirarchy[-i].__dict__.keys() and container_heirarchy[-i].inside == 'parent'):
                        parent = container_heirarchy[-i]
            elif item.inside is not None and len(item.inside) > 0:
                for i in range(1,len(container_heirarchy)+1):
                    if container_heirarchy[-i].name == item.inside:
                        parent = container_heirarchy[-i]
                        break

    return parent

def register_classes(yaml):
    yaml.register_class(UserPath)
    yaml.register_class(Container)
    yaml.register_class(HttpRequest)
    yaml.register_class(Container)
    yaml.register_class(Transaction)
    yaml.register_class(Delay)
    yaml.register_class(RampPolicy)
    yaml.register_class(ConstantPolicy)
    yaml.register_class(Population)
    yaml.register_class(Scenario)
    yaml.register_class(Header)
    yaml.register_class(Extractor)
    yaml.register_class(SLA)
    yaml.register_class(VariableTypeJavascript)
    yaml.register_class(JavascriptAction)


def apply_header_to_user_path(item, user_path):
    apply_header_to_container(item, user_path.init, recurse=True)
    apply_header_to_container(item, user_path.actions, recurse=True)
    apply_header_to_container(item, user_path.end, recurse=True)


def apply_header_to_container(item, container, recurse=False):
    reqs = list(filter(lambda step: type(step) is HttpRequest,container.steps))
    for req in reqs:
        req.headers.append(item)
    if recurse:
        containers = list(filter(lambda step: type(step) is Container or issubclass(type(step), Container), container.steps))
        for c in containers:
            apply_header_to_container(item, c, recurse=recurse)


class MyLogger():
    def __init__(self):
        from io import StringIO
        self.stream = StringIO()

    def write(self, s):
        self.stream.write(s)

    def flush(self):
        self.stream.flush()

    def readAll(self):
        self.stream.seek(0)
        str = self.stream.getvalue()
        self.stream.close()
        return str

# https://stackoverflow.com/questions/55826554/yaml-dumping-a-nested-object-without-types-tags
def strip_python_tags(s):
    result = []
    for line in s.splitlines():
        idx = line.find("!!python/")
        if idx > -1:
            line = line[:idx]
        result.append(line)
    return '\n'.join(result)

# sla http delay transaction delay http sla
#   - does nothing without --apply global
#   - adds a request in the actions container
#   - adds a delay in the actions container
#   - adds a transaction in the actions container
#   - adds a delay in the above transaction
#   - adds a request in the above transaction
#   - applies an sla to the above request
# ramp duration

def write_to_path(path, output):
    dir = path

    if os.path.isfile(path):
        dir = os.path.abspath(os.path.join(path, os.pardir))

    os.makedirs(dir, exist_ok=True)

    if os.path.isdir(path):
        path = os.path.abspath(dir) + os.sep + "default.yaml"

    logging.debug(path)
    logging.debug(output)

    output += "\n"

    with open(path, "w+") as stream:
        stream.write(output)

    return path



__builder_data_singleton = common.__load_data(__builder_config_file)



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

    # extract --name traceId \
    #         --jsonpath ".headers['X-Amzn-Trace-Id']"
    #         --regexp "=(.*)" \

@click.command()
@click.option("--name", type=str, default=None, required=True, help="The name of the extractor/variable to use later")
@click.option("--jsonpath", type=str, default=None, help="The JSON-Path expression to select the data from the response.")
@click.option("--xpath", type=str, default=None, help="The XPath expression to select the data from the response.")
@click.option("--regexp", type=str, default=None, help="The regular expression to select the data from the response.")
@click.option("--from",'from_', type=str, default=None, help="Where to extract the value: 'body' (default), 'header', or 'both'.")
@click.option("--match-number", type=int, default=None, help="The match group to use of regexp produces multiple results.")
@click.option("--template", type=str, default=None, help="The template to construct the value; numbers are match groups from regexp.")
@click.option("--decode", type=str, default=None, help="How to decode the value; default is not to decode; options are 'url' or 'html'.")
@click.option("--extract-once", is_flag=True, default=None, help="Only take the first occurrence of the value extracted.")
@click.option("--default", type=str, default=None, help="The default value; by default is empty.")
@click.option("--throw-error", is_flag=True, default=None, help="Throw an error if no value is extractable; default is true.")
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, name, jsonpath, xpath, regexp, from_, match_number, template, decode, extract_once, default, throw_error):
    """Defines an extractor for the most recent request
    """
    builder_data.register_context(ctx)

    if jsonpath is None and xpath is None and regexp is None:
        raise ValueError("Extractor '{}' must have either jsonpath or xpath or regexp defined.".format(name))

    builder = builder_data.get() \
        .add(builder_data.Extractor(
            name=name,
            jsonpath=jsonpath,
            xpath=xpath,
            regexp=regexp,
            from_=from_,
            match_number=match_number,
            template=template,
            decode=decode,
            extract_once=extract_once,
            default=default,
            throw_error=throw_error)
        ) \
        .save()



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.option('--to', type=int, default=5, help="The max VUs/concurrency to ramp the population")
@click.option('--by', type=int, default=1, help="The # of VUs to add every 'per' interval")
@click.option('--per', default="5s", help="The interval to add VUs 'by'")
@click.option('--duration', help="The duration of this policy")
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, to, by, per, duration):
    """Defines a ramp-up policy for the current User Path
    """
    builder_data.register_context(ctx)

    builder = builder_data.get() \
        .add(builder_data.RampPolicy(to=to, by=by, per=per, duration=duration)) \
        .save()



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.argument("spec", required=False, type=str)
@click.option('--name', '-n', required=False, help="The name of the header")
@click.option('--value', '-v', required=False, help="The value of the header")
@click.option('--all', is_flag=True, help="Apply this header to all prior requests")
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, spec, name, value, all):
    """Defines a header for one or more request
    """
    builder_data.register_context(ctx)

    if name == "-":
        name = None

    if name and not value or value and not name:
        raise ValueError("If you specify either name or value arguments, you must provide both.")

    if spec and (name or value):
        raise ValueError("If you specify a name/value spec, you must specify a '-' for the spec.")

    if spec:
        parts = spec.split("=")
        name = parts[0]
        value = "=".join(parts[1:])

    builder = builder_data.get() \
        .add(builder_data.Header(name=name, value=value, all=all)) \
        .save()



from compose_lib import common
import os

from collections import OrderedDict
from ruamel.yaml.comments import CommentedMap as ordereddict

__profile_config_file = os.path.join(common.__config_dir, "profile.yaml")

def reset():
    global __profile_data_singleton
    __profile_data_singleton = None
    if os.path.exists(__profile_config_file):
        os.remove(__profile_config_file)

def get(throw=True):
    global __profile_data_singleton
    if __profile_data_singleton is None:
        __profile_data_singleton = init()
    return __profile_data_singleton


def init():
    global __profile_data_singleton
    __profile_data_singleton = ProfileData()
    ProfileData.config_file = __profile_config_file
    return __profile_data_singleton

class ProfileData(common.StorableConfig):
    config_file = None

    def __init__(self):
        super(ProfileData, self).__init__()
        self.default_zone = None
        self.default_test_setting = None

    def set_default_zone(self, zone):
        self.default_zone = zone
        return self

    def set_default_test_setting(self, zone):
        self.default_test_setting = zone
        return self

    @classmethod
    def to_yaml(cls,dumper,self):
        #logging.debug(cls.yaml_flow_style)
        data = common.remove_empty(ordereddict(self.__dict__))

        return dumper.represent_data(data)




__profile_data_singleton = common.__load_data(__profile_config_file)



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.option("--confirm", is_flag=True, default=False, help="Must confirm that you want to reset the current builder data")
@click.pass_context
@CommandCategory("Composing Queue Management")
def cli(ctx, confirm):
    """Resets the current builder data; must include --confirm command
    """
    builder_data.register_context(ctx, auto_reset=False)

    if confirm != True:
        raise cli_exception.CliException("You must use the --confirm option")

    builder_data.reset()



import sys
import os

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

def set_global_continue(continuation):
    global __global_continue
    __global_continue = continuation

def get_global_continue():
    return __global_continue



import sys
import click

from compose_lib import builder_data, profile
from compose_lib.command_category import CommandCategory
from commands import config
import tempfile
import subprocess
import yaml
import json
import logging

from compose_lib.common import os_return, find_nlg_app
from threading import Thread, Event
from neoload.neoload_cli_lib import tools

class SubprocessEvent():
    def __init__(self):
        self.underlying_event = Event()
        self.proc = None
        self.stdout = None
        self.stderr = None

    def set_outcomes(self, proc, stdout, stderr):
        self.proc = proc
        self.stdout = stdout
        self.stderr = stderr

    def set(self):
        self.underlying_event.set()

    def is_set(self):
        return self.underlying_event.is_set()

@click.command()
@click.argument("file", required=False)
@click.pass_context
@CommandCategory("Validating")
def cli(ctx, file):
    """Uses the local NeoLoad installation to try a project
    """
    builder_data.register_context(ctx, auto_reset=False)

    yaml_str = None
    if file is None:
        yaml_str = builder_data.convert_builder_to_yaml(builder_data.get())
    else:
        yaml_str = open(file, encoding="UTF-8").read()

    full_path = None
    tmp = tempfile.TemporaryDirectory()
    dir = tmp.name
    full_path = builder_data.write_to_path(dir, yaml_str)

    nlg_app = find_nlg_app()

    event = SubprocessEvent()

    do_it = lambda app=nlg_app, project=full_path, event=event: launch_app(app,project,event)
    action_thread = Thread(target=do_it)

    # Here we start the thread and we wait 5 seconds before the code continues to execute.
    action_thread.start()
    action_thread.join(timeout=20)

    # We send a signal that the other thread should stop.
    event.set()

    msg = "This test project is valid."

    retcode = event.proc.returncode if event.proc is not None else 0
    logging.debug("EXIT_CODE: {}".format(retcode))
    if retcode is not None and retcode != 0:
        if event.stdout is not None:
            msg = event.stdout.decode("UTF-8")
        else:
            msg = "Check console for error details."

    if retcode is None:
        event.proc.kill()

    tools.system_exit({'code': 0 if retcode is None else retcode,'message':msg})
    return


def launch_app(app,project,event):
    proc = os_return("'{}' -project '{}' -exit -noGUI".format(app, project), status=True, connect_stdout=False)
    event.proc = proc
    (stdout,strerr) = proc.communicate()
    event.set_outcomes(proc,stdout,strerr)



import sys
import click

from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

import os

@click.command()
@click.argument("back", type=int, default=1)
@click.pass_context
@CommandCategory("Composing Queue Management")
def cli(ctx, back):
    """Rolls back the prior [back] number of builder commands
    """
    builder_data.register_context(ctx, auto_reset=False)

    data = builder_data.get()
    for i in range(0,back):
        if len(data.stack) > 0:
            data.stack.pop()

    data.save()



import sys
import click

from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.option('--duration', help="The duration of this policy")
@click.argument("vus", type=int)
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, duration, vus):
    """Defines a constant load policy for the current User Path
    """
    builder_data.register_context(ctx)

    builder = builder_data.get() \
        .add(builder_data.ConstantPolicy(vus, duration=duration)) \
        .save()



import appdirs
import os
import jsonpickle
import logging
import colorama

from neoload.neoload_cli_lib import tools,displayer
import subprocess
import asyncio
from datetime import datetime
import html

__conf_name = "neoload-compose"
__version = "1.0"
__author = "neotys"
__config_dir = appdirs.user_data_dir(__conf_name, __author, __version)

class StorableConfig:

    def __init__(self):
        pass

    def save_to_file(self, config_file):
        dir = os.path.abspath(os.path.join(config_file, os.pardir))
        os.makedirs(dir, exist_ok=True)
        with open(config_file, "w") as stream:
            stream.write(jsonpickle.encode(self))
        return self

    def save(self):
        self.save_to_file(type(self).config_file)

    def __str__(self):
        jsonpickle.set_encoder_options('simplejson', sort_keys=True, indent=4)
        jsonpickle.set_encoder_options('json', sort_keys=True, indent=4)
        return jsonpickle.encode(self)

def __load_data(config_file):
    if os.path.exists(config_file):
        with open(config_file, "r") as stream:
            try:
                ret = jsonpickle.decode(stream.read())
                type(ret).config_file = config_file
                return ret
            except Exception:
                logging.warning("Could not load from file {}".format(config_file))

    return None

def remove_empty(od):
    rems = []
    for key in od.keys():
        if od[key] is None: rems.append(key)
    for key in rems:
        del od[key]
    return od


__is_debug = False
def set_debug(debug_on=True):
    global __is_debug
    __is_debug = debug_on


def get_debug():
    global __is_debug
    return __is_debug


def os_return(command, status=False, connect_stdout=True):
    if status:
        print("Running '{}'".format(command))

    actual_command = parse_command(command)

    if connect_stdout:
        p = subprocess.Popen(actual_command, stdout=subprocess.PIPE)
        p.wait()
    else:
        p = subprocess.Popen(actual_command)
    return p

def os_run(command, status=False, print_stdout=False, print_line_check=None):
    if status:
        print("Running '{}'".format(command))
    try:
        if not print_stdout:
            proc = os_return(command, status=False)
            (stdout,strerr) = proc.communicate()
            if proc.returncode != 0:
                tools.system_exit({'code':proc.returncode,'message':"{}".format(strerr)})
                return
        else:
            #subprocess.run(parse_command(command), check = True)
            print(colorama.Style.RESET_ALL)

            asyncio.get_event_loop().run_until_complete(run_async(command, print_line_check))

            print(colorama.Style.RESET_ALL)

        return True
    except subprocess.CalledProcessError as err:
        tools.system_exit({'code':err.returncode,'message':"{}".format(err)})
        return False

async def run_async(cmd, print_line_check=None):
    env = os.environ
    env['PYTHONUNBUFFERED'] = "1"
    p = await asyncio.create_subprocess_shell(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
    if print_line_check is None: print_line_check = lambda x: True
    async for is_out, line in merge(decorate_with(p.stdout, True),
                                    decorate_with(p.stderr, False)):
        if is_out:
            if print_line_check(line.decode('UTF-8').strip()):
                c_print(line.decode('UTF-8').strip(), flush=True)
        else:
            c_print('<text color="Fore.RED">{}</text>'.format(line.decode().strip()), flush=True)

def c_print(*args, flush=True):
    final = "".join(*args)

    final = displayer.colorize_text(final)

    print(final)


async def decorate_with(it, prefix):
    async for item in it:
        yield prefix, item

async def merge(*iterables):
    iter_next = {it.__aiter__(): None for it in iterables}
    while iter_next:
        for it, it_next in iter_next.items():
            if it_next is None:
                fut = asyncio.ensure_future(it.__anext__())
                fut._orig_iter = it
                iter_next[it] = fut
        done, _ = await asyncio.wait(iter_next.values(),
                                     return_when=asyncio.FIRST_COMPLETED)
        for fut in done:
            iter_next[fut._orig_iter] = None
            try:
                ret = fut.result()
            except StopAsyncIteration:
                del iter_next[fut._orig_iter]
                continue
            yield ret

def parse_command(command):
    actual_command = []
    parts = command.split()

    combine = False
    for part in parts:
        if combine:
            actual_command[-1] += " " + part
        else:
            if part.startswith("\"") or part.startswith("'"):
                combine = True

            actual_command.append(part)

        if part.endswith("\"") or part.endswith("'"):
            combine = False

    final_parts = []
    for p in actual_command:
        part = "{}".format(p)
        if part.startswith("\"") or part.startswith("'"):
            part = part[1:]
        if part.endswith("\"") or part.endswith("'"):
            part = part[:-1]
        final_parts.append(part)

    return final_parts

def find_nlg_app():
    return "/Applications/NeoLoad 7.6/bin/NeoLoadGUI.app/Contents/MacOS/NeoLoad"



import sys
import click
import logging
from neoload.neoload_cli_lib import cli_exception
from compose_lib import builder_data
from compose_lib.command_category import CommandCategory

@click.command()
@click.option('--get', help="The URL to use in a GET http request")
@click.option('--post', help="The URL to use in a POST http request")
@click.option('--put', help="The URL to use in a PUT http request")
@click.option('--patch', help="The URL to use in a PATCH http request")
@click.option('--delete', help="The URL to use in a DELETE http request")
@click.option('--option', help="The URL to use in a OPTION http request")
@click.option('--body', help="The body contents to send")
@click.pass_context
@CommandCategory("Composing")
def cli(ctx, get, post, put, patch, delete, option, body):
    """Adds an HTTP request to the builder queue
    """
    builder_data.register_context(ctx)

    details = None
    if details is None: details = process_details("GET", get)
    if details is None: details = process_details("POST", post)
    if details is None: details = process_details("PUT", put)
    if details is None: details = process_details("PATCH", patch)
    if details is None: details = process_details("DELETE", delete)
    if details is None: details = process_details("OPTION", option)

    if details is None:
        raise cli_exception.CliException("You have not provided a proper argument to create the HTTP action from")
    else:
        if body == "-":
            logging.debug("http subcommand using stdin")
            stdin_text = click.get_text_stream('stdin')
            body = stdin_text.read()

        details['body'] = body

    builder = builder_data.get() \
        .add(builder_data.HttpRequest(details)) \
        .save()


def process_details(method, url):
    if url is not None:
        return {
            'method': method,
            'url': url,
            'body': None
        }
    else:
        return None



import logging
import os
import sys

import click
import coloredlogs

from neoload.neoload_cli_lib import cli_exception,tools
from compose_lib import common
from compose_lib.command_category import CommandCategory
from version import __version__

import urllib3

from . import set_global_continue

urllib3.disable_warnings()

plugin_folder = os.path.join(os.path.dirname(__file__), 'commands')

# Disable output buffering.
sys.stdout = sys.__stdout__

def compute_version():
    if __version__ is not None:
        return __version__
    try:
        return os.popen('git describe --tags --dirty').read().strip()
    except (TypeError, ValueError):
        return "dev"


class NeoLoadCompose(click.MultiCommand):

    def format_commands(self, ctx, formatter):
        super().get_usage(ctx)

        formatter.write_paragraph()

        groups = {}
        max_name_len = 10
        for cmd_name in self.list_commands(ctx):
            max_name_len = max(max_name_len,len(cmd_name))
            command = self.get_command(ctx, cmd_name)
            category = "Misc"
            if 'callback' in command.__dict__:
                if 'category' in dir(command.__dict__['callback']):
                    category = command.__dict__['callback'].category
            if category not in groups:
                groups[category] = []
            groups[category].append({
                'name': cmd_name,
                'function': command
            })

        for category in groups:
            commands = groups[category]
            with formatter.section(category):
                for cmd in commands:
                    command = cmd['function']
                    name = cmd['name']
                    formatter.write_text(
                        ('{: <'+str(max_name_len)+'}\t{}').format(name, command.get_short_help_str(limit=(80-max_name_len)))
                        )


    def list_commands(self, ctx):
        """Dynamically get the list of commands."""
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.endswith('.py') and not filename.startswith('__init__'):
                rv.append(filename[:-3].replace('_', '-'))
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Dynamically get the command."""
        ns = {}
        fn = os.path.join(plugin_folder, name.replace('-', '_') + '.py')
        if os.path.isfile(fn):
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
                eval(code, ns, ns)
            return ns['cli']
        else:
            raise cli_exception.CliException("\"" + name + "\" is not a neoload-compose command")

__global_continue = False

@click.command(cls=NeoLoadCompose, help='', chain=True)
@click.option('--debug', default=False, is_flag=True)
@click.option('--continuation', '--continue', '-c', default=False, is_flag=True, help="Append to exiting builder queue, otherwise each separate shell call to this utility resets the builder 'working queue'")
@click.version_option(compute_version())
def cli(debug, continuation):

    common.set_debug(debug)

    if debug:
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True
        cli_exception.CliException.set_debug(True)

    set_global_continue(continuation)

    try:
        if tools.is_color_terminal():
            coloredlogs.install(level=logging.getLogger().level)
    except:
        if sys.stdin.isatty():
            coloredlogs.install(level=logging.getLogger().level)


if __name__ == '__main__':
    cli()
